import bpy

####################
###  MAIN PANEL  ###
####################

class KT_VIEW3D_PT_sync_panel(bpy.types.Panel):
    bl_label = "Sync | Lock Viewport"
    bl_idname = "KT_VIEW3D_PT_sync_panel"
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'UI'
    bl_category = 'K-Tools'
    bl_options = {"DEFAULT_CLOSED"}

    def draw_header(self, context):
        layout = self.layout
        layout.label(text="", icon="UV_SYNC_SELECT")

    def draw(self, context):
        layout = self.layout
        view = context.space_data
        viewLock = view.region_3d
        
        # Apply the active viewport value only once.     
        row = layout.row()
        row.operator("k_tool.sync_operator", text="Match View")
        
        # Clip Start & End
        row = layout.row()
        row.label(text='Clip Start:')
        row.label(text='Clip End:')
        row.label(text='Focal Length:')
        grid = layout.grid_flow(row_major=True, columns=3, align=True)        
        grid.prop(view, "clip_start", text="")
        grid.prop(view, "clip_end", text="")
         
        # Focal Length
        grid.prop(view, "lens", text="")
        
        # Lock view Rotation
        row = layout.row(align=True)
        row.label(text='Lock View Rotation:')
        row = layout.row(align=True)
        row.prop(context.scene, "lock_rotation", text="Lock Active")
        row.prop(context.scene, "k_tool_lock_rotation_all_views", text="Except Active")

        row = layout.row(align=True)                        
        row.label(text="Synchronization:")
        row = layout.row(align=True)
        row.prop(context.scene, "continuous_sync", text="Continuous Sync")


        if context.scene.continuous_sync:
            grid = layout.grid_flow(row_major=True, columns=1, align=True)  
            grid.prop(context.scene, "sync_interval", text="Sync Interval")
            grid.operator("k_tool.sync_interval_operator", text="Start Sync")

################
### SUBPANEL ###
################

class KT_VIEW3D_PT_options_buttons_panel(bpy.types.Panel):
    bl_label = "Options"
    bl_idname = "KT_VIEW3D_PT_options_buttons_panel"
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'UI'
    bl_category = 'K-Tools'
    bl_parent_id = "KT_VIEW3D_PT_sync_panel"
    bl_options = {"DEFAULT_CLOSED"}

    def draw(self, context):
        layout = self.layout
        scene = context.scene
        
        layout.label(text="Sync:")
        
        grid = layout.grid_flow(row_major=False, columns=2, align=True)      

        grid.prop(scene, "sync_view_distance", text="View Distance")
        grid.prop(scene, "sync_view_location", text="View Location")
        grid.prop(scene, "sync_view_rotation", text="View Rotation")
        grid.prop(scene, "sync_camera_zoom", text="Camera Zoom")
        grid.prop(scene, "sync_camera_offset", text="Camera Offset")
        grid.prop(scene, "sync_view_perspective", text="View Perspective")
        grid.prop(scene, "sync_clip_start", text="Clip Start")
        grid.prop(scene, "sync_clip_end", text="Clip End")
        grid.prop(scene, "sync_focal_length", text="Focal Length")
        grid.prop(scene, "sync_all_windows", text="All Windows")
        
        # Add the new button for fine adjustment
        layout.prop(scene, "sync_view_distance_adjust", text="Adjust View Distance")

########################
###  SYNC VIEWPORTS  ###
########################

class KT_VIEW3D_OT_sync_operator(bpy.types.Operator):
    bl_idname = "k_tool.sync_operator"
    bl_label = "Sync Viewport"
    bl_description = "Apply active viewport values only once. For real time sync use 'Continuous Sync'"

    def execute(self, context):
        active_view = context.space_data
        scene = context.scene

        # Determine the areas to sync based on the option buttons
        areas_to_sync = []
        if scene.sync_all_windows:
            for window in bpy.context.window_manager.windows:
                for area in window.screen.areas:
                    if area.type == 'VIEW_3D' and area != context.area:
                        areas_to_sync.append(area)
        else:
            for area in context.screen.areas:
                if area.type == 'VIEW_3D' and area != context.area:
                    areas_to_sync.append(area)

        for area in areas_to_sync:
            for space in area.spaces:
                if space.type == 'VIEW_3D':
                    if scene.sync_view_distance:
                        space.region_3d.view_distance = active_view.region_3d.view_distance * scene.sync_view_distance_adjust
                    if scene.sync_view_location:
                        space.region_3d.view_location = active_view.region_3d.view_location
                    if scene.sync_view_rotation:
                        space.region_3d.view_rotation = active_view.region_3d.view_rotation
                    if scene.sync_camera_zoom:
                        space.region_3d.view_camera_zoom = active_view.region_3d.view_camera_zoom
                    if scene.sync_camera_offset:
                        space.region_3d.view_camera_offset = active_view.region_3d.view_camera_offset
                    if scene.sync_view_perspective:
                        space.region_3d.view_perspective = active_view.region_3d.view_perspective
                    if scene.sync_clip_start:
                        space.clip_start = active_view.clip_start
                    if scene.sync_clip_end:
                        space.clip_end = active_view.clip_end
                    if scene.sync_focal_length:
                        space.lens = active_view.lens
                        
        return {'FINISHED'}

##########################
###  REFRESH INTERVAL  ###
##########################

class KT_VIEW3D_OT_sync_interval_operator(bpy.types.Operator):
    bl_idname = "k_tool.sync_interval_operator"
    bl_label = "Sync Interval Operator"
    bl_description = "Set synchronization automatic"
    
    _timer = None
    _running = False

    action: bpy.props.EnumProperty(
        items=[
            ('START', "Start", "Start synchronization"),
            ('STOP', "Stop", "Stop synchronization"),
        ],
        default='START',
    )

    def modal(self, context, event):
        if not context.scene.continuous_sync:
            self.cancel(context)
            return {'CANCELLED'}

        if event.type == 'TIMER':
            self.execute_sync(context)
            return {'PASS_THROUGH'}

        return {'PASS_THROUGH'}

    def execute(self, context):
        if self.action == 'START' and not self._running:
            wm = context.window_manager
            self._timer = wm.event_timer_add(context.scene.sync_interval, window=context.window)
            wm.modal_handler_add(self)
            self._running = True
            bpy.types.Scene.sync_operator_running = True
        elif self.action == 'STOP' and self._running:
            self.cancel(context)
        return {'RUNNING_MODAL'}

    def cancel(self, context):
        wm = context.window_manager
        if self._timer is not None:
            wm.event_timer_remove(self._timer)
        self._running = False
        bpy.types.Scene.sync_operator_running = False
        return {'CANCELLED'}

    def execute_sync(self, context):
        active_view = context.space_data
        scene = context.scene

        # Determine the areas to sync based on the option buttons
        areas_to_sync = []
        if scene.sync_all_windows:
            for window in bpy.context.window_manager.windows:
                for area in window.screen.areas:
                    if area.type == 'VIEW_3D' and area != context.area:
                        areas_to_sync.append(area)
        else:
            for area in context.screen.areas:
                if area.type == 'VIEW_3D' and area != context.area:
                    areas_to_sync.append(area)

        for area in areas_to_sync:
            for space in area.spaces:
                if space.type == 'VIEW_3D':
                    if scene.sync_view_distance:
                        space.region_3d.view_distance = active_view.region_3d.view_distance * scene.sync_view_distance_adjust
                    if scene.sync_view_location:
                        space.region_3d.view_location = active_view.region_3d.view_location
                    if scene.sync_view_rotation:
                        space.region_3d.view_rotation = active_view.region_3d.view_rotation
                    if scene.sync_camera_zoom:
                        space.region_3d.view_camera_zoom = active_view.region_3d.view_camera_zoom
                    if scene.sync_camera_offset:
                        space.region_3d.view_camera_offset = active_view.region_3d.view_camera_offset
                    if scene.sync_view_perspective:
                        space.region_3d.view_perspective = active_view.region_3d.view_perspective
                    if scene.sync_clip_start:
                        space.clip_start = active_view.clip_start
                    if scene.sync_clip_end:
                        space.clip_end = active_view.clip_end
                    if scene.sync_focal_length:
                        space.lens = active_view.lens

        return {'PASS_THROUGH'}

###########################
### LOCK / UNLOCK VIEW  ###
###########################

def update_lock_rotation(self, context):
    region_3d = context.space_data.region_3d
    region_3d.lock_rotation = self.lock_rotation

bpy.types.Scene.lock_rotation = bpy.props.BoolProperty(
    name="Lock Rotation",
    description="Lock view rotation of side views on Top/Front/Right",
    default=False,
    update=update_lock_rotation
)


class KT_VIEW3D_OT_Toggle_Lock_Active_View(bpy.types.Operator):
    bl_idname = 'k_tools.toggle_lock_active_view'
    bl_label = 'Toggle Lock Rotation'
    
    def execute(self, context):
        context.scene.lock_rotation = not context.scene.lock_rotation
        return {'FINISHED'}


def lock_rotation(context):  
    for area in bpy.context.screen.areas:
        if area.type == 'VIEW_3D' and area != context.area:
            for space in area.spaces:
                if space.type == 'VIEW_3D':
                    space.region_3d.lock_rotation = True

def unlock_rotation(context):
    for area in bpy.context.screen.areas:
        if area.type == 'VIEW_3D':
            for space in area.spaces:
                if space.type == 'VIEW_3D':
                    space.region_3d.lock_rotation = False

class KT_VIEW3D_OT_lock_rotation_operator(bpy.types.Operator):
    bl_idname = "k_tool.lock_rotation_operator"
    bl_label = "Lock Rotation Other Views"

    def execute(self, context):
        if context.scene.k_tool_lock_rotation_all_views:
            lock_rotation(context)
        else:
            unlock_rotation(context)
        return {'FINISHED'}

def lock_rotation_all_views_update(self, context):
    if self.k_tool_lock_rotation_all_views:
        lock_rotation(context)
    else:
        unlock_rotation(context)

########################################
###      REGISTER / UNREGISTER       ###
########################################

classes = (
    KT_VIEW3D_PT_sync_panel,
    KT_VIEW3D_PT_options_buttons_panel,
    KT_VIEW3D_OT_sync_operator,
    KT_VIEW3D_OT_sync_interval_operator,
    KT_VIEW3D_OT_lock_rotation_operator,
    KT_VIEW3D_OT_Toggle_Lock_Active_View,
)    

def register():
    for cls in classes:
        bpy.utils.register_class(cls)

    # Main Panel Buttons
    bpy.types.Scene.continuous_sync = bpy.props.BoolProperty(name="Continuous Sync", default=False, description='Set the synchronization automatic based in a time value')
    bpy.types.Scene.sync_interval = bpy.props.FloatProperty(name="Sync Interval", default=1.0, max=1.0, min=0.01, description='Time value for synchronization, lower = faster')
    bpy.types.Scene.k_tool_lock_rotation_all_views = bpy.props.BoolProperty(name="Lock Rotation All Views", default=False, update=lock_rotation_all_views_update, description="Lock the Rotation for all viewports except the active one")
    
    # Subpanel Buttons - Boolean for sync options
    bpy.types.Scene.sync_view_distance = bpy.props.BoolProperty(name="Sync View Distance", default=True, description="Sync viewing distance between View Distance")
    bpy.types.Scene.sync_view_location = bpy.props.BoolProperty(name="Sync View Location", default=True, description="Sync viewing location between 3d Views")
    bpy.types.Scene.sync_view_rotation = bpy.props.BoolProperty(name="Sync View Rotation", default=True, description="Sync viewing rotation between 3d Views")
    bpy.types.Scene.sync_camera_zoom = bpy.props.BoolProperty(name="Sync Camera Zoom", default=True, description="Sync camera zoom between 3d Views")
    bpy.types.Scene.sync_camera_offset = bpy.props.BoolProperty(name="Sync Camera Offset", default=True, description="Sync camera offset between 3d Views")
    bpy.types.Scene.sync_view_perspective = bpy.props.BoolProperty(name="Sync View Perspective", default=True, description="Sync viewing perspective between 3d Views")
    bpy.types.Scene.sync_clip_start = bpy.props.BoolProperty(name="Sync Clip Start", default=True, description="Sync the start point of the viewing clip between 3d Views")
    bpy.types.Scene.sync_clip_end = bpy.props.BoolProperty(name="Sync Clip End", default=True, description="Sync the end point of the viewing clip between 3d Views")
    bpy.types.Scene.sync_focal_length = bpy.props.BoolProperty(name="Sync Focal Length", default=True, description="Sync focal length between 3d Views")
    bpy.types.Scene.sync_all_windows = bpy.props.BoolProperty(name="Sync All Windows", default=False, description="Sync viewports across all open windows")

    # Add the float property for fine adjustment
    bpy.types.Scene.sync_view_distance_adjust = bpy.props.FloatProperty(name="Adjust View Distance:", default=1.0,  min=0.01, max=5.0, description="Fine adjustment (Multiplier) for Sync View Distance. Lower = Closer | 1 = Default Distance")

def unregister():
    for cls in classes:
        bpy.utils.unregister_class(cls)
        
    # Removing Main Panel Buttons
    del bpy.types.Scene.continuous_sync
    del bpy.types.Scene.sync_interval
    del bpy.types.Scene.k_tool_lock_rotation_all_views

    # Removing Subpanel Buttons - Boolean for sync options
    del bpy.types.Scene.sync_view_distance
    del bpy.types.Scene.sync_view_location
    del bpy.types.Scene.sync_view_rotation
    del bpy.types.Scene.sync_camera_zoom
    del bpy.types.Scene.sync_camera_offset
    del bpy.types.Scene.sync_view_perspective
    del bpy.types.Scene.sync_clip_start
    del bpy.types.Scene.sync_clip_end
    del bpy.types.Scene.sync_focal_length
    del bpy.types.Scene.sync_all_windows
    
    # Remove the float property for fine adjustment
    del bpy.types.Scene.sync_view_distance_adjust
    
if __name__ == "__main__":
    register()